#include "MESH_MeshX3DBuilderCoreVisitor.h"
#include "MESH_Scene.h"
#include "MESH_Transform.h"
#include "MESH_Switch.h"
#include "X3D_Scene.h"
#include "X3D_Transform.h"
#include "X3D_Switch.h"
#include "X3D_Shape.h"
#include "X3D_IndexedFaceSet.h"
#include "X3D_Coordinate.h"
#include "X3D_X3DAppearanceNode.h"
#include "X3D_Normal.h"
#include "X3D_ColorRGBA.h"
#include "X3D_TextureCoordinate.h"

using namespace std;

namespace X3DTK {
namespace MESH {

MeshX3DBuilderCoreVisitor::MeshX3DBuilderCoreVisitor()
{
  defineEnterFunction(&MeshX3DBuilderCoreVisitor::enterScene);
  defineEnterFunction(&MeshX3DBuilderCoreVisitor::enterTransform);
  defineEnterFunction(&MeshX3DBuilderCoreVisitor::enterSwitch);
  defineEnterFunction(&MeshX3DBuilderCoreVisitor::enterMesh);
  defineEnterFunction(&MeshX3DBuilderCoreVisitor::enterVertexSet);
  
  defineLeaveFunction(&MeshX3DBuilderCoreVisitor::leaveX3DNode);
  defineLeaveFunction(&MeshX3DBuilderCoreVisitor::leaveMesh);
  
  stateVariables = GraphTraversal::getInstanceOf<MeshX3DBuilderStateVariables>();
}

void MeshX3DBuilderCoreVisitor::enterScene(Scene *S) const
{
//  cout << "enterScene" << endl;
  X3D::Scene *NS = dynamic_cast<X3D::Scene *>(stateVariables->getNode(S));
  if (NS == 0)
  {
    NS = new X3D::Scene();
    // Updating Scene.
    X3D::MFNode x3dChildren = S->getX3DNodeChildren();
    for (X3D::MFNode::const_iterator it = x3dChildren.begin(); it != x3dChildren.end(); ++it)
      NS->addChild(stateVariables->clone(*it));
    
    stateVariables->addCoupleNode(S, NS);
  }
  stateVariables->pushNode(NS); 
}

void MeshX3DBuilderCoreVisitor::enterTransform(Transform *T) const
{
//  cout << "enterTransform" << endl;

  X3D::Transform *NT = dynamic_cast<X3D::Transform *>(stateVariables->getNode(T));
  if (NT == 0)
  {
    NT = new X3D::Transform();
    // Updating Transform.
    NT->setCenter(T->getCenter());
    NT->setRotation(T->getRotation());
    NT->setScale(T->getScale());
    NT->setScaleOrientation(T->getScaleOrientation());
    NT->setTranslation(T->getTranslation());
    
    X3D::MFNode x3dChildren = T->getX3DNodeChildren();
    for (X3D::MFNode::const_iterator it = x3dChildren.begin(); it != x3dChildren.end(); ++it)
      NT->addChild(stateVariables->clone(*it));
    
    stateVariables->addCoupleNode(T, NT);
  }
  stateVariables->pushNode(NT); 
}

void MeshX3DBuilderCoreVisitor::enterSwitch(Switch *S) const
{
  X3D::Switch *NS = dynamic_cast<X3D::Switch *>(stateVariables->getNode(S));
  if (NS == 0)
  {
    NS = new X3D::Switch();
    // Updating Switch.
    NS->setWhichChoice(S->getWhichChoice());
    
    X3D::MFNode x3dChildren = S->getX3DNodeChildren();
    for (X3D::MFNode::const_iterator it = x3dChildren.begin(); it != x3dChildren.end(); ++it)
      NS->addChild(stateVariables->clone(*it));
    
    stateVariables->addCoupleNode(S, NS);
  }
  stateVariables->pushNode(NS); 
}

void MeshX3DBuilderCoreVisitor::enterMesh(Mesh *M) const
{
  X3D::Shape *NS = dynamic_cast<X3D::Shape *>(stateVariables->getNode(M));
  X3D::X3DGeometryNode *NN = 0;
  if (NS == 0)
  {
    NS = new X3D::Shape();
    // Updating Shape.
    NS->setAppearance(dynamic_cast<X3D::X3DAppearanceNode *>(stateVariables->clone(M->getX3DAppearanceNodeChild())));
    X3D::IndexedFaceSet *NI = new X3D::IndexedFaceSet();
    
    // Updating IndexedFaceSet.
    // setting the parameters.
    NI->setCcw(true);
    NI->setNormalPerVertex(M->data().getNormalPerVertex());
    NI->setNormalPerVertex(M->data().getColorPerVertex());
    
    // Vertices
    const MESH::Mesh::MFVertex &mv = M->getVertices();
    // Faces
    MFInt32 coordIndexArray;
    const MESH::Mesh::MFFace &mf = M->getFaces(); 
    
    // Filling the coordIndex.
    for (MESH::Mesh::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
    {
      const MESH::Mesh::MFEdge &me = (*itFace)->getEdges();
      for (MESH::Mesh::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
        coordIndexArray.push_back((*itEdge)->getFromVertex()->getIndex());
      coordIndexArray.push_back(-1);
    }
  
    NI->setCoordIndex(coordIndexArray);
    
    ///////////////////////////////////////////////////////////////////////////////////
    // normals
  
    if (M->data().hasNormal())
    {
      X3D::Normal *N = new X3D::Normal();
      NI->setNormal(N);
      
      MFVec3f normalArray;

      if (M->data().getNormalPerVertex())
      {
        // iterating the faces to create the normalIndexArray.
        MFInt32 normalIndexArray;
      
        for (MESH::Mesh::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        {
          const MESH::Mesh::MFEdge &me = (*itFace)->getEdges();
          for (MESH::Mesh::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
          {
            MESH::Mesh::SFVertex *V = (*itEdge)->getFromVertex();
            const SFVec3f &normal = V->data().getNormalOfFace(*itFace);
            
            unsigned int i = 0;
            for (MFVec3f::const_iterator itNormal = normalArray.begin(); itNormal != normalArray.end(); ++itNormal)
            {
              if (*itNormal == normal)
                break;
              ++i;  
            }
            if (i == normalArray.size())
              normalArray.push_back(normal);
            normalIndexArray.push_back(i);
          }  
          normalIndexArray.push_back(-1);
        }
        NI->setNormalIndex(normalIndexArray);
      }
      else
      {
        for (MESH::Mesh::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
          normalArray.push_back((*itFace)->data().getNormal());
      }
      N->setVector(normalArray);
    }
    
    //////////////////////////////////////////////////////////////////////////////////// 
    // colors
  
    if (M->data().hasColor())
    {
      // creating a new Node.
      X3D::ColorRGBA *C = new X3D::ColorRGBA();
      NI->setColor(C);
  
      MFColorRGBA colorArray;

      if (M->data().getColorPerVertex())
      {
        // iterating the faces to create the colorIndexArray.
        MFInt32 colorIndexArray;
      
        for (MESH::Mesh::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        {
          const MESH::Mesh::MFEdge &me = (*itFace)->getEdges();
          for (MESH::Mesh::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
          {
            MESH::Mesh::SFVertex *V = (*itEdge)->getFromVertex();
            const SFColorRGBA &color = V->data().getColorOfFace(*itFace);
            
            unsigned int i = 0;
            for (MFColorRGBA::const_iterator itColor = colorArray.begin(); itColor != colorArray.end(); ++itColor)
            {
              if (*itColor == color)
                break;
              ++i;  
            }
            if (i == colorArray.size())
              colorArray.push_back(color);
            colorIndexArray.push_back(i);
          }  
          colorIndexArray.push_back(-1);
        }
        NI->setColorIndex(colorIndexArray);
      }
      else
      {
        for (MESH::Mesh::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
          colorArray.push_back((*itFace)->data().getColor());
      }
      C->setColor(colorArray);
    }
 
    //////////////////////////////////////////////////////////////////////////////////// 
    // texCoords
  
    if (M->data().hasTexCoord())
    {
      X3D::TextureCoordinate *T = new X3D::TextureCoordinate();
      NI->setTexCoord(T);
    
      MFVec2f texCoordArray;
      // iterating the faces to create the normalIndexArray.
      MFInt32 texCoordIndexArray;
    
      for (MESH::Mesh::MFFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
      {
        const MESH::Mesh::MFEdge &me = (*itFace)->getEdges();
        for (MESH::Mesh::MFEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
        {
          MESH::Mesh::SFVertex *V = (*itEdge)->getFromVertex();
          const SFVec2f &texCoord = V->data().getTexCoordOfFace(*itFace);
          
          unsigned int i = 0;
          for (MFVec2f::const_iterator itTexCoord = texCoordArray.begin(); itTexCoord != texCoordArray.end(); ++itTexCoord)
          {
            if (*itTexCoord == texCoord)
              break;
            ++i;  
          }
          if (i == texCoordArray.size())
            texCoordArray.push_back(texCoord);
          texCoordIndexArray.push_back(i);
        }  
        texCoordIndexArray.push_back(-1);
      }
      NI->setTexCoordIndex(texCoordIndexArray);
      T->setPoint(texCoordArray);
    }
    
    // ending the links.
    NS->setGeometry(NI);
    NN = NI;
    
    stateVariables->addCoupleNode(M, NS);
  }
  else
    NN = static_cast<X3D::X3DGeometryNode *>(NS->getGeometry());

  stateVariables->pushNode(NS);
  stateVariables->pushNode(NN);
}

void MeshX3DBuilderCoreVisitor::enterVertexSet(VertexSet *V) const
{
  X3D::Coordinate *NC = dynamic_cast<X3D::Coordinate *>(stateVariables->getNode(V));
  if (NC == 0)
  {
    NC = new X3D::Coordinate();
    // Updating Coordinate.
    const Mesh::MFVertex &mv = V->getVertices();
    MFVec3f point;
    for (Mesh::MFVertex::const_iterator it = mv.begin(); it != mv.end(); ++it)
      point.push_back(SFVec3f((*it)->data().getPoint()));
    NC->setPoint(point);
    
    stateVariables->addCoupleNode(V, NC);
  }
  stateVariables->pushNode(NC); 
}

void MeshX3DBuilderCoreVisitor::leaveX3DNode(X3DNode *N) const
{
//  cout << "leaveX3DNode " << N->getTypeName() << endl;
  stateVariables->popNode();
}

void MeshX3DBuilderCoreVisitor::leaveMesh(Mesh *) const
{
//  cout << "leaveMesh" << endl;
  stateVariables->popNode();
  stateVariables->popNode();
}

} 
} 
